#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ogcsys.h>
#include <ogc/lwp_watchdog.h>
#include <unistd.h> //for sleep

#include "apploader.h"
#include "disc.h"
#include "subsystem.h"
#include "video.h"
#include "wdvd.h"
#include "fat.h"
#include "selectdol.h"

#define MIGHTY_CHEATS			"fat:/config/mighty/cheats/"

/* Constants */
#define PTABLE_OFFSET	0x40000
#define WII_MAGIC	0x5D1C9EA3

/* Disc pointers */
static u32 *buffer = (u32 *)0x93000000;
static u8  *diskid = (u8  *)0x80000000;


void __Disc_SetLowMem(void)
{
	/* Setup low memory */
	*(vu32 *)0x80000030 = 0x00000000;
	*(vu32 *)0x80000060 = 0x38A00040;
	*(vu32 *)0x800000E4 = 0x80431A80;
	*(vu32 *)0x800000EC = 0x81800000;
	*(vu32 *)0x800000F4 = 0x817E5480;
	*(vu32 *)0x800000F8 = 0x0E7BE2C0;
	*(vu32 *)0x800000FC = 0x2B73A840;

	/* Copy disc ID */
	// online check code, seems offline games clear it?
	memcpy((void *)0x80003180, (void *)0x80000000, 4);

	// Fix for Sam & Max (WiiPower)
	//*(vu32*)0x80003184	= 0x80000000;	// Game ID Address

	/* Flush cache */
	DCFlushRange((void *)0x80000000, 0x3F00);
}

void __Disc_SetVMode(u8 videomode)
{
	GXRModeObj *vmode = NULL;

	u32 progressive, tvmode, vmode_reg = 0;

	/* Get video mode configuration */
	progressive = (CONF_GetProgressiveScan() > 0) && VIDEO_HaveComponentCable();
	tvmode      =  CONF_GetVideo();

	/* Select video mode register */
	switch (tvmode) {
	case CONF_VIDEO_PAL:
		if (CONF_GetEuRGB60() > 0) {
			vmode_reg = 5;
			vmode     = (progressive) ? &TVNtsc480Prog : &TVEurgb60Hz480IntDf;
		} else
			vmode_reg = 1;

		break;

	case CONF_VIDEO_MPAL:
		vmode_reg = 4;
		break;

	case CONF_VIDEO_NTSC:
		vmode_reg = 0;
		break;
	}

	if(videomode==0){
		/* Select video mode */
		switch(diskid[3]) {
		/* PAL */
		case 'D':
		case 'F':
		case 'P':
		case 'X':
		case 'Y':
			if (tvmode != CONF_VIDEO_PAL) {
				vmode_reg = 1;
				vmode     = (progressive) ? &TVNtsc480Prog : &TVNtsc480IntDf;
			}
	
			break;
	
		/* NTSC or unknown */
		case 'E':
		case 'J':
			if (tvmode != CONF_VIDEO_NTSC) {
				vmode_reg = 0;
				vmode     = (progressive) ? &TVNtsc480Prog : &TVEurgb60Hz480IntDf;
			}
	
			break;
		}
	}else if(videomode==1){
		if (CONF_GetEuRGB60() > 0) {
			vmode_reg = 5;
		}else{
			vmode_reg = 1;
		}
		vmode     = (progressive) ? &TVNtsc480Prog : &TVEurgb60Hz480IntDf;
	}else if(videomode==2){
		vmode_reg = 1;
		vmode     = (progressive) ? &TVNtsc480Prog : &TVNtsc480IntDf;
	}

	/* Set video mode register */
	*(vu32 *)0x800000CC = vmode_reg;

	/* Set video mode */
	if (vmode)
		Video_Configure(vmode);

	/* Clear screen */
	Video_Clear(COLOR_BLACK);
}

void __Disc_SetTime(void)
{
	/* Extern */
	//extern void settime(long long);

	/* Set proper time */
	settime(secs_to_ticks(time(NULL) - 946684800));
}

s32 __Disc_FindPartition(u64 *outbuf)
{
	u64 offset = 0, table_offset = 0;

	u32 cnt, nb_partitions;
	s32 ret;

	/* Read partition info */
	ret = WDVD_UnencryptedRead(buffer, 0x20, PTABLE_OFFSET);
	if (ret < 0)
		return ret;

	/* Get data */
	nb_partitions = buffer[0];
	table_offset  = buffer[1] << 2;

	/* Read partition table */
	ret = WDVD_UnencryptedRead(buffer, 0x20, table_offset);
	if (ret < 0)
		return ret;

	/* Find game partition */
	for (cnt = 0; cnt < nb_partitions; cnt++) {
		u32 type = buffer[cnt * 2 + 1];

		/* Game partition */
		if(!type)
			offset = buffer[cnt * 2] << 2;
	}

	/* No game partition found */
	if (!offset)
		return -1;

	/* Set output buffer */
	*outbuf = offset;

	return 0;
}


s32 Disc_Init(void)
{
	/* Init DVD subsystem */
	return WDVD_Init();
}

s32 Disc_Open(void)
{
	s32 ret;

	/* Reset drive */
	ret = WDVD_Reset();
	if (ret < 0)
		return ret;

	/* Read disc ID */
	return WDVD_ReadDiskId(diskid);
}

s32 Disc_Wait(void)
{
	u32 cover = 0;
	s32 ret;
	int timeout=0;

	/* Wait for disc */
	while (!(cover & 0x2)) {
		/* Get cover status */
		ret = WDVD_GetCoverStatus(&cover);
		if (ret < 0)
			return ret;

		timeout++;
		sleep(1);
		if(timeout==20)
			return -1;
	}

	return 0;
}

s32 Disc_SetWBFS(u32 mode, u8 *id)
{
	/* Set WBFS mode */
	return WDVD_SetWBFSMode(mode, id);
}

s32 Disc_ReadHeader(void *outbuf)
{
	/* Read disc header */
	return WDVD_UnencryptedRead(outbuf, sizeof(struct discHdr), 0);
}

s32 Disc_IsWii(void)
{
	struct discHdr *header = (struct discHdr *)buffer;

	s32 ret;

	/* Read disc header */
	ret = Disc_ReadHeader(header);
	if (ret < 0)
		return ret;

	/* Check magic word */
	if (header->magic != WII_MAGIC)
		return -1;

	return 0;
}

s32 Disc_BootPartition(u64 offset, u8 modoVideo, bool anti002, u8 ocarina, u16 altdol){
	entry_point p_entry;

	s32 ret;

	/* Open specified partition */
	ret = WDVD_OpenPartition(offset);
	if (ret < 0)
		return ret;

	/* Run apploader */
	ret = Apploader_Run(&p_entry,(ocarina>0), altdol);
	if (ret < 0)
		return ret;

	// Remove 002
	if(anti002)
    	*(u32 *)0x80003140 = *(u32 *)0x80003188;

    
	/* Setup low memory */
	__Disc_SetLowMem();

	/* Set an appropiate video mode */
	__Disc_SetVMode(modoVideo);

	/* Set time */
	__Disc_SetTime();

	// Ocarina
    if(ocarina){
	    /* OCARINA STUFF - FISHEARS */
	    static char gameid[8];
	    memset(gameid, 0, 8);
		memcpy(gameid, (char*)0x80000000, 6);

		void *filebuff;

		char filepath[64];
		if(ocarina==1)
			sprintf(filepath, MIGHTY_CHEATS "%s.gct", gameid);
		else
			sprintf(filepath, MIGHTY_CHEATS "%s_%d.gct", gameid, ocarina);

		ret = Fat_ReadFile(filepath, &filebuff);

		if(ret>0){
			//printf("* SD Codes found (%d bytes).\n", ret);
			//sleep(2);
			memcpy((void*)0x800027E8, filebuff, ret);
			*(vu8*)0x80001807 = 0x01;
	
		}
	    /* OCARINA STUFF - FISHEARS*/
    }
	Fat_Unmount();
    
    /* Close subsystems */
	Subsystem_Close();

	/* Shutdown IOS */
	SYS_ResetSystem(SYS_SHUTDOWN, 0, 0);

	/* SAM AND MAX PARAMETERS - originally from tueidj */
	if(0x80000000=='R' && 0x80000000=='3' && (0x80000000=='X' || 0x80000000=='Y')){
		//R3X - sam and max
		//R3Y - sam and max 2

		/*
		param=17; //tutorial
		param=25; //episodio 1
		param=33; //episodio 2
		param=41; //episodio 3
		param=49; //episodio 4
		param=57; //episodio 5
		param=65; //episodio 6 (ste no existe para sam and max 2)
		*(u32*)0xCC003024 = param; //Originally from tueidj
		*/
	}




	/* METROID PRIME TRILOGY fail PARAMETERS */
	//*(u32*)0x8016A604 = 0x00000000;
	//*(u32*)0x8016A604 = 0x00000000|1;
	//*(u32*)0x8096a338 = not tested;
	//__asm__ ( "lis %r3, -0x7fe9; addi %r3, %r3, -0x59FC\nli %r4, 0x1\nstw %r4, 0(%r3)\n" );
	//__asm__ ( "lis %r3, -0x7fe9; addi %r3, %r3, -0x59FC\nli %r4, 0\nstw %r4, 0(%r3)\n" ); //no va
	//__asm__ ( "lis %r3, 0x8016\nori %r3, %r3, 0xa604\nli %r4, 0\nstw %r4, 0(%r3)\n" ); //no va
	//__asm__ ( "lis %r3, 0x8096\nori %r3, %r3, 0xa604\nli %r4, 0\nstw %r4, 0(%r3)\n" ); //no va
	//*(u32*)0x8016a338 = 0x7C7E1B00; //malo
	//*(u32*)0x8016A338 = 0x00000000; //NTSC???
	//*(u32*)0x8096a338=0; //no va

	/* Jump to entry point */
	p_entry();

	/* Epic failure */
	while (1);

	return 0;
}

s32 Disc_WiiBoot(u8 modoVideo,bool anti002,u8 ocarina,u16 altdol){
	u64 offset;
	s32 ret;

	/* Find game partition offset */
	ret = __Disc_FindPartition(&offset);
	if (ret < 0)
		return ret;

	/* Boot partition */
	return Disc_BootPartition(offset,modoVideo,anti002,ocarina,altdol);
}
